﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Forms;

using Midi;

namespace RB3KB_USB2MIDI {
    public partial class FormMain : Form {

        private static OutputDevice MIDI_OUT = null;
        private static Channel channel = Channel.Channel1;
        private static USBKB keyboard = null;

        enum PedalMode { Expression, ChannelVolume, FootController };

        Thread thread = null;
        private int Octave = 4;
        private int Program = 0;

        #region Set Octave
        public void SetOctave(int octave) {
            octave = Math.Min(octave, 8);
            octave = Math.Max(octave, 0);
            Octave = octave;

            this.Invoke((MethodInvoker)delegate {
                txtOctave.Text = Octave.ToString();
            });
        }

        private void txtOctave_Leave(object sender, EventArgs e) {
            int octave;
            if (Int32.TryParse(txtOctave.Text, out octave))
                SetOctave(octave);
        }

        private void txtOctave_KeyPress(object sender, KeyPressEventArgs e) {
            if (e.KeyChar == (char)Keys.Enter)
                txtOctave_Leave(sender, e);
        }
        #endregion

        #region Set Program / Instrument
        public void SetProgram(int program) {
            program = Math.Min(program, 127);
            program = Math.Max(program, 0);
            Program = program;

            if(MIDI_OUT != null && MIDI_OUT.IsOpen)
                MIDI_OUT.SendProgramChange(channel, (Instrument)Program);

            this.Invoke((MethodInvoker)delegate {
                txtInstrumentNum.Text = (Program + 1).ToString();
                ddlInstrument.SelectedIndex = Program;
            });
        }

        private void ddlInstrument_SelectedIndexChanged(object sender, EventArgs e) {
            SetProgram(ddlInstrument.SelectedIndex);
        }

        private void txtInstrumentNum_Leave(object sender, EventArgs e) {
            int program;
            if (Int32.TryParse(txtInstrumentNum.Text, out program))
                SetProgram(program - 1);
        }

        private void txtInstrumentNum_KeyPress(object sender, KeyPressEventArgs e) {
            if (e.KeyChar == (char)Keys.Enter)
                txtInstrumentNum_Leave(sender, e);
        }
        #endregion

        public FormMain() {
            InitializeComponent();
        }

        private void FormMain_Load(object sender, EventArgs e) {
            foreach (OutputDevice device in OutputDevice.InstalledDevices) {
                ddlMidiOut.Items.Add(device.Name);

                if (device.Name == Properties.Settings.Default.LastUsedMidiOut)
                    ddlMidiOut.SelectedIndex = ddlMidiOut.Items.Count - 1;
            }

            foreach (string name in Enum.GetNames(typeof(Instrument)))
                ddlInstrument.Items.Add(name);

            SetOctave(4);
            SetProgram(0);
        }

        private void Start() {
            MIDI_OUT.Open();
            keyboard = new USBKB();
        }

        private void End() {
            if(keyboard != null)
                keyboard.Close();

            if(MIDI_OUT != null && MIDI_OUT.IsOpen)
                MIDI_OUT.Close();
        }

        private void USB2MIDI() {
            keyboard.SetLED(1, true);

            //bool DrumMapping = false;

            bool holdPanic = false;
            bool holdProgram = false;
            bool holdOctave = false;

            MIDI_OUT.SendProgramChange(channel, (Instrument)Program);

            Dictionary<int, int> HeldKey = new Dictionary<int, int>();

            KBData previous = keyboard.Read();
            while (true) {
                KBData data = keyboard.Read();

                //Not ready
                //if (data.DpadUp) DrumMapping = !DrumMapping;

                #region Panic: All Notes Off (Select, Home and Start together)
                if(data.SelectMinus && data.Home && data.StartPlus && !holdPanic) {
                    holdPanic = true;
                    int max = (8 * 12) + 25;
                    for (int i = 0; i < max; i++)
                        MIDI_OUT.SendNoteOff(channel, (Pitch)i, 0);
                } else if (!data.SelectMinus && !data.Home && !data.StartPlus && holdPanic) {
                    holdPanic = false;
                }
                #endregion

                #region Program
                if (data.Triangle2 && !holdProgram) {
                    holdProgram = true;
                    SetProgram(Program+1);
                    keyboard.LEDAnimateLTR();
                } else if (data.CrossA && !holdProgram) {
                    holdProgram = true;
                    SetProgram(Program-1);
                    keyboard.LEDAnimateRTL();
                } else if (!data.Triangle2 && !data.CrossA && holdProgram) {
                    holdProgram = false;
                }
                #endregion

                #region Octave
                if (data.CircleB && !holdOctave) {
                    holdOctave = true;
                    SetOctave(Octave+1);
                    keyboard.LEDAnimateLTR();
                } else if (data.Square1 && !holdOctave) {
                    holdOctave = true;
                    SetOctave(Octave-1);
                    keyboard.LEDAnimateRTL();
                } else if (!data.CircleB && !data.Square1 && holdOctave) {
                    holdOctave = false;
                }
                #endregion

                #region Keys
                for (int i = 0; i < data.Key.Length; i++) {
                    if (data.Key[i] && !previous.Key[i]) {
                        int pitch = (Octave * 12) + i;

                        //Probably only correct when only pressing keys further right of already pressed keys.
                        int velocity = 80;
                        for(int v = data.Velocity.Length - 1; v >= 0; v--) {
                            if(data.Velocity[v] != 0) {
                                velocity = data.Velocity[v];
                                break;
                            }
                        }

                        HeldKey.Add(i, pitch);
                        MIDI_OUT.SendNoteOn(channel, (Pitch)pitch, velocity);
                    } else if (!data.Key[i] && previous.Key[i] && HeldKey.ContainsKey(i)) {
                        MIDI_OUT.SendNoteOff(channel, (Pitch)HeldKey[i], 0);
                        HeldKey.Remove(i);
                    }
                }
                #endregion

                #region Slider and Overdrive (Pitch Bend and Modulation)
                if (data.Slider != previous.Slider) {
                    if(data.Overdrive) {
                        MIDI_OUT.SendPitchBend(channel, (data.Slider * 140));
                    } else {
                        MIDI_OUT.SendControlChange(channel, Midi.Control.ModulationWheel, data.Slider);
                    }
                }
                #endregion

                previous = data;
            }
        }

        private void ddlMidiOut_SelectedIndexChanged(object sender, EventArgs e) {
            MIDI_OUT = OutputDevice.InstalledDevices[ddlMidiOut.SelectedIndex];

            Properties.Settings.Default.LastUsedMidiOut = ddlMidiOut.SelectedItem.ToString();
            Properties.Settings.Default.Save();
        }

        private void btnRunLoop_Click(object sender, EventArgs e) {
            if (thread == null || !thread.IsAlive) {
                if (MIDI_OUT != null) {
                    Start();

                    if (keyboard.CanUse()) {
                        thread = new Thread(() => USB2MIDI());
                        thread.Start();

                        ddlMidiOut.Enabled = false;
                        btnRunLoop.Text = "Running";
                    } else {
                        End();
                        ddlMidiOut.Enabled = true;
                        btnRunLoop.Text = "Keyboard not found - missing filter?";
                    }
                }
            } else {
                thread.Abort();
                End();

                ddlMidiOut.Enabled = true;
                btnRunLoop.Text = "Doing Nothing";
            }
        }

        private void FormMain_FormClosing(object sender, FormClosingEventArgs e) {
            if(thread != null)
                thread.Abort();
            End();
        }

    }
}
